from typing import Mapping, Union

import pyarrow

class Bool:
    def ddl(self) -> str | None: ...

class Bytea:
    def ddl(self) -> str | None: ...

class Int8:
    def ddl(self) -> str | None: ...

class Int2:
    def ddl(self) -> str | None: ...

class Int4:
    def ddl(self) -> str | None: ...

class Char:
    def ddl(self) -> str | None: ...

class Text:
    def ddl(self) -> str | None: ...

class Jsonb:
    def ddl(self) -> str | None: ...

class Float4:
    def ddl(self) -> str | None: ...

class Float8:
    def ddl(self) -> str | None: ...

class Date:
    def ddl(self) -> str | None: ...

class Time:
    def ddl(self) -> str | None: ...

class Timestamp:
    def ddl(self) -> str | None: ...

class Interval:
    def ddl(self) -> str | None: ...

class List:
    def __init__(self, __type: Column) -> None: ...
    def ddl(self) -> str | None: ...

class Column:
    def __init__(self, __nullable: bool, __type: PostgresType) -> None: ...
    @property
    def data_type(self) -> PostgresType: ...
    @property
    def nullable(self) -> bool: ...

class PostgresSchema:
    def __init__(self, columns: list[tuple[str, Column]]) -> None: ...
    @property
    def columns(self) -> list[tuple[str, Column]]: ...

PostgresType = Union[
    Bool,
    Bytea,
    Int2,
    Int4,
    Int8,
    Float4,
    Float8,
    Char,
    Text,
    Jsonb,
    Date,
    Time,
    Timestamp,
    Interval,
    List,
]

class ArrowToPostgresBinaryEncoder:
    def __init__(self, __schema: pyarrow.Schema) -> None: ...
    @staticmethod
    def new_with_encoders(
        __schema: pyarrow.Schema, __encoders: Mapping[str, EncoderBuilder]
    ) -> ArrowToPostgresBinaryEncoder: ...
    def write_header(self) -> bytes: ...
    def write_batch(self, __batch: pyarrow.RecordBatch) -> bytes: ...
    def finish(self) -> bytes: ...
    def schema(self) -> PostgresSchema: ...
    @staticmethod
    def infer_encoder(__field: pyarrow.Field) -> EncoderBuilder: ...

class BooleanEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class UInt8EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class UInt16EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class UInt32EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Int8EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...
    @classmethod
    def new_with_output(
        cls, field: pyarrow.Field, output: Char | Int2
    ) -> Int8EncoderBuilder: ...

class Int16EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Int32EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Int64EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Float16EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Float32EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Float64EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class TimestampMicrosecondEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class TimestampMillisecondEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class TimestampSecondEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Date32EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Date64EncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Time32MillisecondEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Time32SecondEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class Time64MicrosecondEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class DurationMicrosecondEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class DurationMillisecondEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class DurationSecondEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class StringEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...
    @classmethod
    def new_with_output(
        cls, field: pyarrow.Field, output: Text | Jsonb
    ) -> Int8EncoderBuilder: ...

class LargeStringEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...
    @classmethod
    def new_with_output(
        cls, field: pyarrow.Field, output: Text | Jsonb
    ) -> Int8EncoderBuilder: ...

class BinaryEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class LargeBinaryEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...

class ListEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...
    @classmethod
    def new_with_inner(
        cls, field: pyarrow.Field, inner_encoder_builder: EncoderBuilder
    ) -> ListEncoderBuilder: ...

class LargeListEncoderBuilder:
    def __init__(self, field: pyarrow.Field) -> None: ...
    @classmethod
    def new_with_inner(
        cls, field: pyarrow.Field, inner_encoder_builder: EncoderBuilder
    ) -> LargeListEncoderBuilder: ...

EncoderBuilder = (
    BooleanEncoderBuilder
    | UInt8EncoderBuilder
    | UInt16EncoderBuilder
    | UInt32EncoderBuilder
    | Int8EncoderBuilder
    | Int16EncoderBuilder
    | Int32EncoderBuilder
    | Int64EncoderBuilder
    | Float16EncoderBuilder
    | Float32EncoderBuilder
    | Float64EncoderBuilder
    | TimestampMicrosecondEncoderBuilder
    | TimestampMillisecondEncoderBuilder
    | TimestampSecondEncoderBuilder
    | Date32EncoderBuilder
    | Date64EncoderBuilder
    | Time32MillisecondEncoderBuilder
    | Time32SecondEncoderBuilder
    | Time64MicrosecondEncoderBuilder
    | DurationMicrosecondEncoderBuilder
    | DurationMillisecondEncoderBuilder
    | DurationSecondEncoderBuilder
    | StringEncoderBuilder
    | LargeStringEncoderBuilder
    | BinaryEncoderBuilder
    | LargeBinaryEncoderBuilder
    | ListEncoderBuilder
    | LargeListEncoderBuilder
)



PYCLASS_TEMPLATE = """\
#[pyclass(module = "pgpq._pgpq")]
struct {name} {{
    inner: encoders::EncoderBuilder,
}}
field_passthrough_impl!({name});
"""

CLASS_TEMPLATE = """\
class {name}:
    def __init__(self, field: pyarrow.Field) -> None:
        ...
"""

field_only = [
    "BooleanEncoderBuilder",
    "UInt8EncoderBuilder",
    "UInt16EncoderBuilder",
    "UInt32EncoderBuilder",
    "Int8EncoderBuilder",
    "Int16EncoderBuilder",
    "Int32EncoderBuilder",
    "Int64EncoderBuilder",
    "Float16EncoderBuilder",
    "Float32EncoderBuilder",
    "Float64EncoderBuilder",
    "TimestampMicrosecondEncoderBuilder",
    "TimestampMillisecondEncoderBuilder",
    "TimestampSecondEncoderBuilder",
    "Date32EncoderBuilder",
    "Date64EncoderBuilder",
    "Time32MillisecondEncoderBuilder",
    "Time32SecondEncoderBuilder",
    "Time64MicrosecondEncoderBuilder",
    "DurationMicrosecondEncoderBuilder",
    "DurationMillisecondEncoderBuilder",
    "DurationSecondEncoderBuilder",
    "StringEncoderBuilder",
    "LargeStringEncoderBuilder",
    "BinaryEncoderBuilder",
    "LargeBinaryEncoderBuilder",
    "ListEncoderBuilder",
    "LargeListEncoderBuilder",
    "DynamicEncoderBuilder",
]

types = [
    "Bool",
    "Bytea",
    "Int8",
    "Int2",
    "Int4",
    "Char",
    "Text",
    "Jsonb",
    "Float4",
    "Float8",
    "Date",
    "Time",
    "Timestamp",
    "Interval",
]

TYPE_TEMPLATE = """\
#[pyclass(module = "pgpq._pgpq")]
struct {name} {{
    inner: PostgresType,
}}
impl_simple!({name}, PostgresType::{name});
"""

PY_TYPE_TEMPLATE = """\
class {name}:
    pass
"""

for name in field_only:
    print(PYCLASS_TEMPLATE.format(name=name))

for name in field_only:
    print(CLASS_TEMPLATE.format(name=name))

for name in types:
    print(TYPE_TEMPLATE.format(name=name))

for name in types:
    print(PY_TYPE_TEMPLATE.format(name=name))



from pgpq._pgpq import (
    BooleanEncoderBuilder,
    UInt8EncoderBuilder,
    UInt16EncoderBuilder,
    UInt32EncoderBuilder,
    Int8EncoderBuilder,
    Int16EncoderBuilder,
    Int32EncoderBuilder,
    Int64EncoderBuilder,
    Float16EncoderBuilder,
    Float32EncoderBuilder,
    Float64EncoderBuilder,
    TimestampMicrosecondEncoderBuilder,
    TimestampMillisecondEncoderBuilder,
    TimestampSecondEncoderBuilder,
    Date32EncoderBuilder,
    Date64EncoderBuilder,
    Time32MillisecondEncoderBuilder,
    Time32SecondEncoderBuilder,
    Time64MicrosecondEncoderBuilder,
    DurationMicrosecondEncoderBuilder,
    DurationMillisecondEncoderBuilder,
    DurationSecondEncoderBuilder,
    StringEncoderBuilder,
    LargeStringEncoderBuilder,
    BinaryEncoderBuilder,
    LargeBinaryEncoderBuilder,
    ListEncoderBuilder,
    LargeListEncoderBuilder,
)

__all__ = (
    "BooleanEncoderBuilder",
    "UInt8EncoderBuilder",
    "UInt16EncoderBuilder",
    "UInt32EncoderBuilder",
    "Int8EncoderBuilder",
    "Int16EncoderBuilder",
    "Int32EncoderBuilder",
    "Int64EncoderBuilder",
    "Float16EncoderBuilder",
    "Float32EncoderBuilder",
    "Float64EncoderBuilder",
    "TimestampMicrosecondEncoderBuilder",
    "TimestampMillisecondEncoderBuilder",
    "TimestampSecondEncoderBuilder",
    "Date32EncoderBuilder",
    "Date64EncoderBuilder",
    "Time32MillisecondEncoderBuilder",
    "Time32SecondEncoderBuilder",
    "Time64MicrosecondEncoderBuilder",
    "DurationMicrosecondEncoderBuilder",
    "DurationMillisecondEncoderBuilder",
    "DurationSecondEncoderBuilder",
    "StringEncoderBuilder",
    "LargeStringEncoderBuilder",
    "BinaryEncoderBuilder",
    "LargeBinaryEncoderBuilder",
    "ListEncoderBuilder",
    "LargeListEncoderBuilder",
)



from __future__ import annotations

from glob import glob
from pathlib import Path
from typing import Any, Iterator, List, Tuple

import pyarrow as pa
import pyarrow.ipc as paipc
import pytest
from testing.postgresql import Postgresql
import psycopg

import pgpq.schema
import pgpq.encoders
from pgpq import ArrowToPostgresBinaryEncoder
from pgpq.schema import PostgresSchema


@pytest.fixture(scope="session")
def postgres():
    return Postgresql()


Connection = psycopg.Connection[Tuple[Any, ...]]


@pytest.fixture(scope="session")
def dbconn(postgres: Postgresql) -> Iterator[Connection]:
    print(postgres.url())
    with psycopg.connect(str(postgres.url())) as conn:
        yield conn


def copy_buffer_and_get_rows(
    schema: PostgresSchema, buffer: bytes, dbconn: Connection
) -> List[Tuple[Any, ...]]:
    cols = [f'"{col_name}" {col.data_type.ddl()}' for col_name, col in schema.columns]
    ddl = f"CREATE TEMP TABLE data ({','.join(cols)})"
    try:
        with dbconn.cursor() as cursor:
            cursor.execute(ddl)  # type: ignore
            with cursor.copy("COPY data FROM STDIN WITH (FORMAT BINARY)") as copy:
                copy.write(buffer)
            cursor.execute("SELECT * FROM data")
            rows = cursor.fetchall()
    finally:
        dbconn.rollback()  # all that matters is that we drop our temp table
    return rows


TESTCASES = [
    f.strip(".bin").split("/")[-1] for f in sorted(glob("core/tests/snapshots/*"))
]


@pytest.mark.parametrize(
    "testcase",
    TESTCASES,
)
def test_encode_record_batch(dbconn: Connection, testcase: str) -> None:
    path = Path("core/tests/testdata") / f"{testcase}.arrow"
    arrow_table = paipc.open_file(path).read_all()
    encoder = ArrowToPostgresBinaryEncoder(arrow_table.schema)
    buffer = bytearray()
    buffer.extend(encoder.write_header())
    for batch in arrow_table.to_batches():
        buffer.extend(encoder.write_batch(batch))
    buffer.extend(encoder.finish())

    pg_schema = encoder.schema()

    rows = copy_buffer_and_get_rows(pg_schema, buffer, dbconn)
    col = arrow_table.schema.names[0]
    new_table = pa.Table.from_pylist(
        [{col: row[0]} for row in rows], schema=arrow_table.schema
    )

    assert arrow_table == new_table


def test_schema(dbconn: Connection) -> None:
    batch = pa.RecordBatch.from_arrays(
        [
            pa.array([0, 1, 2]),
            pa.array([True, False, None]),
            pa.array([None, ["foo"], ["bar"]]),
            pa.array([[""], ["foo", None], ["bar"]]),
        ],
        schema=pa.schema(
            [
                ("int", pa.int32()),
                ("nullable bool", pa.bool_()),
                pa.field(
                    "a nullable list of strings",
                    pa.list_(pa.field("field", pa.string(), nullable=False)),
                    nullable=True,
                ),
                pa.field(
                    "a list of nullable strings",
                    pa.list_(pa.field("field", pa.string(), nullable=True)),
                    nullable=False,
                ),
            ]
        ),
    )

    encoder = ArrowToPostgresBinaryEncoder(batch.schema)
    schema = encoder.schema()

    assert schema == pgpq.schema.PostgresSchema(
        [
            ("int", pgpq.schema.Column(True, pgpq.schema.Int4())),
            ("nullable bool", pgpq.schema.Column(True, pgpq.schema.Bool())),
            (
                "a nullable list of strings",
                pgpq.schema.Column(
                    True,
                    pgpq.schema.List(
                        pgpq.schema.Column(
                            False,
                            pgpq.schema.Text(),
                        )
                    ),
                ),
            ),
            (
                "a list of nullable strings",
                pgpq.schema.Column(
                    False,
                    pgpq.schema.List(
                        pgpq.schema.Column(
                            True,
                            pgpq.schema.Text(),
                        )
                    ),
                ),
            ),
        ]
    )


def test_infer_encoder() -> None:
    batch = pa.RecordBatch.from_arrays(
        [
            pa.array([0, 1, 2]),
            pa.array([True, False, None]),
            pa.array([None, ["foo"], ["bar"]]),
            pa.array([[""], ["foo", None], ["bar"]]),
        ],
        schema=pa.schema(
            [
                ("int", pa.int32()),
                ("nullable bool", pa.bool_()),
                pa.field(
                    "a nullable list of strings",
                    pa.list_(pa.field("field", pa.string(), nullable=False)),
                    nullable=True,
                ),
                pa.field(
                    "a list of nullable strings",
                    pa.list_(pa.field("field", pa.string(), nullable=True)),
                    nullable=False,
                ),
            ]
        ),
    )

    encoders = {
        field_name: ArrowToPostgresBinaryEncoder.infer_encoder(
            batch.schema.field(field_name)
        )
        for field_name in batch.schema.names
    }

    assert encoders == {
        "int": pgpq.encoders.Int32EncoderBuilder(batch.schema.field("int")),
        "nullable bool": pgpq.encoders.BooleanEncoderBuilder(
            batch.schema.field("nullable bool")
        ),
        "a nullable list of strings": pgpq.encoders.ListEncoderBuilder.new_with_inner(
            batch.schema.field("a nullable list of strings"),
            pgpq.encoders.StringEncoderBuilder(
                batch.schema.field("a nullable list of strings").type.value_field
            ),
        ),
        "a list of nullable strings": pgpq.encoders.ListEncoderBuilder.new_with_inner(
            batch.schema.field(
                "a list of nullable strings",
            ),
            pgpq.encoders.StringEncoderBuilder(
                batch.schema.field("a list of nullable strings").type.value_field
            ),
        ),
    }


def test_custom_encoding(dbconn: Connection) -> None:
    batch = pa.RecordBatch.from_arrays(
        [
            pa.array(
                [["[]"], ['{"foo":"bar"}'], ["123"]],
                type=pa.list_(pa.field("field", pa.string())),
            ),
        ],
        schema=pa.schema(
            [
                pa.field(
                    "json_list",
                    pa.list_(pa.field("field", pa.string())),
                ),
            ]
        ),
    )

    encoders = {
        "json_list": pgpq.encoders.ListEncoderBuilder.new_with_inner(
            batch.schema.field("json_list"),
            pgpq.encoders.StringEncoderBuilder.new_with_output(
                batch.schema.field("json_list").type.value_field, pgpq.schema.Text()
            ),
        )
    }

    encoder = ArrowToPostgresBinaryEncoder.new_with_encoders(batch.schema, encoders)
    buffer = bytearray()
    buffer.extend(encoder.write_header())
    buffer.extend(encoder.write_batch(batch))
    buffer.extend(encoder.finish())

    pg_schema = encoder.schema()

    rows = copy_buffer_and_get_rows(pg_schema, buffer, dbconn)
    print(rows)



"""Example for README.md"""
import psycopg
import pyarrow.dataset as ds
from pgpq import ArrowToPostgresBinaryEncoder


dataset = ds.dataset("tests/testdata/")

encoder = ArrowToPostgresBinaryEncoder(dataset.schema)
# get the expected postgres destination schema
# note that this is _not_ the same as the incoming arrow schema
# and not necessarily the schema of your permanent table
# instead it's the schema of the data that will be sent over the wire
# which for example does not have timezones on any timestamps
pg_schema = encoder.schema()
# assemble ddl for a temporary table
# it's often a good idea to bulk load into a temp table to:
# (1) Avoid indexes
# (2) Stay in-memory as long as possible
# (3) Be more flexible with types (you can't load a SMALLINT into a BIGINT col)
cols = [f"\"{col['name']}\" {col['data_type']['ddl']}" for col in pg_schema["columns"]]
ddl = f"CREATE TEMP TABLE data ({','.join(cols)})"

with psycopg.connect("postgres://postgres:postgres@localhost:5432/postgres") as conn:
    with conn.cursor() as cursor:
        cursor.execute(ddl)  # type: ignore
        with cursor.copy("COPY data FROM STDIN WITH (FORMAT BINARY)") as copy:
            copy.write(encoder.write_header())
            for batch in dataset.to_batches():
                copy.write(encoder.write_batch(batch))
            copy.write(encoder.finish())
        # load into your actual table, possibly doing type casts
        # cursor.execute("INSERT INTO \"table\" SELECT * FROM data")



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pgpq._pgpq import PostgresType as PostgresType


from pgpq._pgpq import (
    Bool,
    Bytea,
    Int8,
    Int2,
    Int4,
    Char,
    Text,
    Jsonb,
    Float4,
    Float8,
    Date,
    Time,
    Timestamp,
    Interval,
    List,
    Column,
    PostgresSchema,
)


__all__ = (
    "Bool",
    "Bytea",
    "Int8",
    "Int2",
    "Int4",
    "Char",
    "Text",
    "Jsonb",
    "Float4",
    "Float8",
    "Date",
    "Time",
    "Timestamp",
    "Interval",
    "List",
    "Column",
    "PostgresSchema",
)
